/*
 * Copyright 2000 by Jean-Claude Christophe <jch@oleane.net>
 *                                          <jch@bsdfr.org>
 *
 * Home page:  http://www.bsdfr.org/projects/tac_bsd/
 *
 * packet.c  This file contains functions for build packets
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>

#include "draft2.h"
#include "tac_bsd.h"
#include "conf.h"

t_pkt_hdr *build_header(u_char, int);

extern int send_data(void *, int, int);
extern int read_data(void *, int, int);
extern int md5_xor(t_pkt_hdr *, u_char *, char *);

int check_header(t_pkt_hdr *, int);

int
send_authen(sd, user, password, port)
    int  sd;        /* socket descriptor */
    char *user;     /* login name */
    char *password; /* password of the user */
    char *port;     /* port used by the user (it makes sens on logging) */
{
    t_pkt_hdr *myhead;    /* packet's header */
    t_send_authen mybody; /* packet's body */
    int res;
    void *pkt;
    int pkt_len;
    int datalength = AUTHEN_START_SIZE;
    
    port = port + 5; /* we don't need the '/dev/' */
    
    myhead = build_header(TAC_PLUS_AUTHEN, TAC_PLUS_UNENCRYPTED_FLAG);
    
    /* it's time to fill the packet's body */
    mybody.action       = TAC_PLUS_AUTHEN_LOGIN;
    mybody.priv_lvl     = TAC_PLUS_PRIV_LVL_MIN;
    mybody.authen_type  = TAC_PLUS_AUTHEN_TYPE_ASCII;
    mybody.service      = TAC_PLUS_AUTHEN_SVC_LOGIN;
    mybody.user_len     = strlen(user);
    mybody.port_len     = strlen(port);
    mybody.rem_addr_len = mybody.data_len = 0;
    myhead->seq_no = seq_no = 1; /* first time we send a packet; */
    myhead->session_id  = session_id = (int)random();
    myhead->version = (TAC_PLUS_MAJOR_VER | TAC_PLUS_MINOR_VER_DEFAULT);
    
    /* building the packet */
    pkt = (u_char *)malloc(pkt_len = HEADER_SIZE + AUTHEN_START_SIZE + \
			   mybody.user_len + mybody.port_len + mybody.rem_addr_len + \
			   mybody.data_len);
    
    /* username */
    bcopy(user, pkt + datalength + HEADER_SIZE, mybody.user_len);
    datalength += mybody.user_len;
    
    /* port */
    bcopy(port, pkt + datalength + HEADER_SIZE, mybody.port_len);
    datalength += mybody.port_len;
    
    /* rem_addr: we don't need it (not required) */
    /* password: we don't need it (not used for now) */
    
    myhead->length = htonl(datalength);
    
    bcopy(myhead, pkt, HEADER_SIZE);
    bcopy(&mybody, pkt + HEADER_SIZE, AUTHEN_START_SIZE);
    
    if (secret)
	md5_xor(pkt, pkt + HEADER_SIZE, secret);
    
    /* The packet is now complete, now we have to write it in our socket (sd). */
    res = send_data(pkt, pkt_len, sd); /* we write only the header for now: 12 bytes */
    
    if ((res < 0) && slog) {
	syslog(_syslog, "in %s, packet header too short: expected %d", __FUNCTION__, HEADER_SIZE);
	return(-1);
    }
    
    free(pkt);
    free(myhead);
    
    return(0);
}

int read_authen(sd) 
    int sd;
{
    t_pkt_hdr myhead;
    t_authen_reply *mybody;
    int len_from_header, res;

    /* read the reply header */
    res = read_data(&myhead, HEADER_SIZE, sd);
    
    if (res < 0) {
	if (slog)
	    syslog(_syslog, "in %s, received header with size different of %d", \
		   __FUNCTION__, HEADER_SIZE);
	return(-1);
    }
    
    /* good header ? */
    if (check_header(&myhead, TAC_PLUS_AUTHEN) < 0) {
	if (slog)
	    syslog(_syslog, "in %s, received header with type %d, expected %d", \
		   __FUNCTION__, myhead.type, TAC_PLUS_AUTHEN);
	return(-1);
    }
    
    len_from_header = ntohl(myhead.length);
    mybody = (struct s_authen_reply *)malloc(len_from_header);
    
    /* read reply packet body */
    res = read_data(mybody, len_from_header, sd);
    
    if (res < 0) {
	if (slog)
	    syslog(_syslog, "in %s, received body with size different of %d", \
		   __FUNCTION__, len_from_header);
	return(-1);
    }

    free(mybody);
    
    return(0);
}

t_pkt_hdr *build_header(type, enc)
    u_char type; /* type of packet we wanna send */
    char enc;    /* encrypted or not */
{
    t_pkt_hdr *header;
    
    header = (struct s_pkt_hdr *)malloc(HEADER_SIZE); /* memory for header */
    
    /*
     * ok we have a struct who defines our header, we can return it now,
     * but it's better to fill it with some information before :-)
     */
    
    /* 
     * we cannot define version here because it will depend on request
     * we cannot define the seq_no here because it will change each time
     * we'll send a packet (normally seq_no += 2)
     */
    
    header->type = type;
    header->flags = enc;
    /* same problem for session_id, we'll set it later */
    
    return(header);
}

int
check_header(hdr, type)
    t_pkt_hdr *hdr;
    int        type;
{
    /* we have 3 things to check: type, seq_no and session_id */
    
    /* First, the type */
    if (hdr->type != type) {
	if (slog)
	    syslog(_syslog, "in %s, received type %d, expected %d", \
		   __FUNCTION__, hdr->type, type);
	return(-1);
    }
    
    /* Second, the seq_no */
    if (hdr->seq_no != seq_no + 1) {
	if (slog)
	    syslog(_syslog, "in %s, incorrect sequence number, received %d, " \
		   "expected %d", __FUNCTION__, hdr->seq_no, seq_no + 1);
	return(-1);
    }
    
    /* Third, the session_id */
    if (hdr->session_id != session_id) {
	if (slog)
	    syslog(_syslog, "in %s, incorrect session number, received %d, " \
		   "expected %d", __FUNCTION__, hdr->session_id, session_id);
	return(-1);
    }
    return(0);
}

int
send_auth_cont(sd, msg, msg_len)
    int  sd;      /* socket descriptor, normally, the session is always open */
    char *msg;
    int  msg_len;
{
    t_pkt_hdr *myhead;
    t_cont_authen mybody;
    void* pkt;
    int pkt_len, res;
    
    pkt_len = HEADER_SIZE + AUTHEN_CONT_SIZE + msg_len;
    pkt = (void *)malloc(pkt_len);

    if (!pkt) { /* not enough memory */
	perror("pkt");
	exit(-1);
    }
    
    /* building and filling the header.. */
    myhead = build_header(TAC_PLUS_AUTHEN, TAC_PLUS_UNENCRYPTED_FLAG);
    myhead->version = (TAC_PLUS_MAJOR_VER | TAC_PLUS_MINOR_VER_DEFAULT);
    myhead->seq_no = seq_no + 2; /* the response IS a sequence. */
    seq_no = myhead->seq_no;     /* it will be useful later for checking the header */
    myhead->length = htonl(msg_len + AUTHEN_CONT_SIZE);
    
    /* building the body.. */
    mybody.user_data_len = mybody.flags = 0;
    mybody.user_msg_len = htons(msg_len);
    bcopy(msg, pkt + HEADER_SIZE + AUTHEN_CONT_SIZE, msg_len); 
    bcopy(myhead, pkt, HEADER_SIZE);
    bcopy(&mybody, pkt + HEADER_SIZE, AUTHEN_CONT_SIZE);
    
    /* crypting.. */
    if (secret)
	md5_xor(pkt, pkt + HEADER_SIZE, secret);
    
    /* sending.. */
    res = send_data(pkt, pkt_len, sd);
    if (res < 0) {
	if (slog)
	    syslog(_syslog, "in %s, packet body too short: expected %d", \
		   __FUNCTION__, pkt_len);
	return(-1);
    }
    
    /* freeing.. */
    free(pkt);
    
    return(0);
}

int
read_result(sd)
    int sd;
{
    t_pkt_hdr myhead;
    t_authen_reply *mybody;
    int len_from_header, res;
    
    /* read the reply header */
    res = read(sd, &myhead, HEADER_SIZE);
    
    if (res < HEADER_SIZE) {
	if (slog)
	    syslog(_syslog, "in %s, received header with size %d, expected %d", \
		   __FUNCTION__, res, HEADER_SIZE);
	return(-1);
    }
    
    /* good header ? */
    if (check_header(&myhead, TAC_PLUS_AUTHEN) < 0) {
	if (slog)
	    syslog(_syslog, "in %s, received header with type %d, expected %d", \
		   __FUNCTION__, myhead.type, TAC_PLUS_AUTHEN);
	return(-1);
    }
    
    len_from_header = ntohl(myhead.length);
    mybody = (struct s_authen_reply *)malloc(len_from_header);
    
    /* read reply packet body */
    res = read(sd, mybody, len_from_header);
    if (res < len_from_header) {
	printf("Probleme3\n");
	return(-1);
    }
    
    if (secret)
	md5_xor(&myhead, (void *)mybody, secret);
    
    if (mybody->status == TAC_PLUS_AUTHEN_STATUS_PASS) /* We won */
	return(0);
    
    return(-1); /* We lost, or unexpected response from server */
}
